# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

from   ophac.dtypes import *
import ophac.ultrametric as ult


def _getLogger(x):
    import logging
    return logging.getLogger(__name__ + '.' + x.__name__)

def HACUntied(lnk):
    log = _getLogger(HACUntied)    
    cpp = True
    try:
        import ophac_cpp
        log.info('C++ extension available.')
    except ModuleNotFoundError as e:
        log.warning('C++ extension not available.')
        cpp = False
    
    if cpp:
        log.info('Using C++ extension.')
        return HACUntied_cpp(lnk)
    else:
        log.info('Using python implementation.')
        return HACUntied_python(lnk)

class HACUntied_cpp:
       
    def __init__(self,lnk):
        self.log     = _getLogger(HACUntied_cpp)
        self.lnk     = lnk
        self.log.info('Instantiated with L:%s', lnk)

    def generate(self,dissim,order=None,mode='untied',**kwargs):
        import time
        import ophac_cpp    as cpp
        import numpy        as np
        import numpy.random as rnd

        CPP_MAX_UINT = 4294967294

        if 'seed' in kwargs:
            self.seed_used = cpp.seed(kwargs['seed'])
        else:
            self.seed_used = cpp.seed(int(rnd.randint(CPP_MAX_UINT)))
        
        self.log.info('Running on %d element space with mode %s and seed %d.',
                      dissim.n, mode, self.seed_used)

        dd = dissim.dists
        qq = [list() for _ in np.arange(dissim.n)]        
        if order:
            qq = order.quivers
            
        cpp_start = time.time()
        ac = None
        if mode == 'approx':
            ac = cpp.linkage_approx(dd,qq,self.lnk)
        else:
            assert mode == 'untied'
            ac = cpp.linkage_untied(dd,qq,self.lnk)

        self.log.info('C++ ophac (%s) completed in %1.3f s.',
                      mode, time.time() - cpp_start)

        dists,joins = zip(*ac)
        return AC(dists=dists,joins=joins)

            
        
class HACUntied_python:

    def __init__(self, lnk):
        '''
        lnk  - Linkage model; one of 'single', 'average' or 'complete'.
        '''
        self.log    = _getLogger(HACUntied_python)
        self.lnk    = lnk
        self._setLinkageFunctionFactory()
        self.log.info('Instantiated with lnk=%s.', lnk)

    def _setLinkageFunctionFactory(self):
        lnkFact = None
        if self.lnk == 'single':
            lnkFact = lambda sizes : BasicLinkage(SingleLinkage())
        elif self.lnk == 'complete':
            lnkFact = lambda sizes : BasicLinkage(CompleteLinkage())
        elif self.lnk == 'average':
            lnkFact = lambda sizes : BasicLinkage(AverageLinkage(sizes))
        else:
            raise Exception('Unknown linkage: "%s"' % self.lnk)

        self._getLinkageFunction = lnkFact

    def generate(self,dissim,order=None,mode='untied'):
        '''
        dissim - Dissimilarity measure of type ophac.dtypes.DistMatrix 
        order  - Order relation of type ophac.dtypes.Quivers, 
                 or None for non-ordered clustering.
        '''
        if mode != 'untied':
            raise Exception('Only untied mode is available in python.')
        
        if order is None:
            order = Quivers(n=dissim.n, relation=[])

        P0   = Partition(n=dissim.n)        
        return self._exploreChains(dissim, order, P0, AC())
        
    def _exploreChains(self, dissim, order, partition, ac0):
        '''
        dissim    - dissimilarity measure for the current partition
        order     - Quivers object for the current partition
        partition - Current partition
        ac0       - AgglomerativeClustering object leading up to the 
                    current partition
        '''
        self.log.debug('Exploring set of %d elements.', len(partition))

        if len(order) == 1:
            self.log.debug('Trivial partition reached.')
            return ac0
    
        # Find minimal dissimilarity level to merge on
        dispair = None
        for (d,pair) in dissim.getSortedIndexPairs():
            if dispair is not None:
                break # Has one!

            a,b = pair
            if order.canMerge(a,b):
                self.log.debug('Mergeable pair (%d,%d) of dissim %1.3f found.',
                               a, b, dissim[a,b])
                dispair = (d,pair)

        if dispair is None:
            self.log.debug('No further mergeable pairs --- maximal element reached')
            return ac0
        
        # Get current linkage
        linker = self._getLinkageFunction([len(x) for x in partition])
        
        (dist,(a,b)) = dispair
        self.log.debug('Merging (%d,%d) at dissim %1.3f.', a, b, dist)
        D2  = linker(a,b,dissim)
        P2  = partition.merge(a,b)
        O2  = order.merge(a,b)
        ac2 = ac0 + AC(joins=[(a,b)], dists=[dist])
        if len(ac2) > 1:
            assert ac2.dists[-2] <= ac2.dists[-1]

        return self._exploreChains(D2, O2, P2, ac2)




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

'''
Plotting support for partial dendrograms
Plotting requires SciPy. Install with `pip install ophac[plot]` or just `pip install scipy`.
'''

def plot(ac,N,*args,**kwargs):
    '''
    Plots the (partial) dendrogram using
    scipy.cluster.hierarchy.dendrogram.
    Note: the caller must call matplotlib.pyplot.show()
    after calling this method for the actual window to show up.
    ac - Thea AgglomerativeClustering to display
    N  - The number of elements in the clustering
    args and kwargs : additional parameters passed on to ...show(...)
    '''
    try:
        import scipy.cluster.hierarchy as sch
    except ImportError:
        raise ImportError("Plotting requires SciPy. Install with `pip install ophac[plot]`.")

    Z,cutDist = _toLinkageMatrix(ac,N)

    myargs = {
        'color_threshold' : cutDist,
        'above_threshold_color' : 'none',
        }

    myargs.update(kwargs)

    sch.dendrogram(Z=Z, *args, **myargs)

def _toLinkageMatrix(ac,N,eps=1e-12):
    '''
    Generates a linkage matrix that is suitable for plotting
    of dendrograms through scipy.cluster.hierarchy.dendrogram
    ac - AgglomerativeClustering object returned by HC.
    N  - Number of elements being clustered
    returns (Z,L,Ncc)
    Z       - The linkage matrix
    cutDist - The distance used to render the 'fake' links to complete
              the matrix.
    '''
    import numpy as np

    L = len(ac)
    Z = np.zeros((N-1,4))

    # Mapping from cluster index in agg to cluster sequential index
    idxMap = list(range(N))

    # Sequential index of the next cluster to be formed
    idx = N

    # Cluster sizes in ac indices
    sizes = [1]*N

    # Compute the partial diagram merges
    for i in range(L):
        join = ac.joins[i]
        dist = ac.dists[i]

        a,b    = (idxMap[join[0]], idxMap[join[1]])
        size   = _mergeSizes(*join,sizes=sizes)
        Z[i,:] = [a,b,dist,size]

        _updateIndices(*join,new=idx,idxMap=idxMap)
        idx += 1

    # Merge the rest at the last distance level plus an epsilon
    # to avoid over-writing lines
    dist = Z[L-1,2] + eps
    for i in range(L,N-1):
        join   = (len(sizes)-2,len(sizes)-1)
        a,b    = (idxMap[join[0]], idxMap[join[1]])
        size   = _mergeSizes(*join,sizes=sizes)
        Z[i,:] = [a,b,dist,size]
        _updateIndices(*join,new=idx,idxMap=idxMap)
        idx += 1

    return (Z,ac.dists[-1] + eps)

def _updateIndices(a,b,new,idxMap):
    idxMap[a] = new
    for i in range(b,len(idxMap)-1):
        idxMap[i] = idxMap[i+1]

    del idxMap[-1]

def _mergeSizes(a,b,sizes):
    assert a < b
    sizes[a] += sizes[b]
    for i in range(b,len(sizes)-1):
        sizes[i] = sizes[i+1]

    del sizes[-1]

    return sizes[a]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


'''
This init file defines the containing folder to be a namespace package,
it allows for re-opening.

This file was auomatically generated at Wed Feb 5 13:42:06 CET 2020
'''

from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from . import _version
__version__ = _version.__version__

def test():
    '''
    Smoke test of installation.
    '''
    from . import rnd, hierarchy
    N = 30
    dissimMatrix,quivers  = rnd.randomOrderedDissimSpace(N,0.1,3)
    D = dissimMatrix.dists
    Q = quivers.quivers
    assert hierarchy.linkage(D,Q)        is not None
    assert hierarchy.approx_linkage(D,Q) is not None
    print('Smoke test of ophac seems to work just fine.')



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

import sys
import re

class ArgSyntaxError(Exception):
    def __init__(self, msg):
        super(ArgSyntaxError, self).__init__('Command line argument syntax error - ' + msg)


class args:
    def __init__(self, opts='', kws={}, defaults={}, help={}):
        self._kwTypeMap = kws
        self._opts      = opts
        self._defaults  = defaults
        self._initOpts(opts)
        self._initAttributes()
        self._helpTexts = help
        assert set(help.keys()).issubset(kws.keys())
        
    def parse(self, argv=sys.argv[1:]):
        extr = re.compile(r'(^--(?P<key>\w+)([:](?P<val>.+))?$)|(^-(?P<opts>\w+)$)')

        keysToSet = set(self._kwTypeMap.keys()) - set(self._defaults.keys())

        for arg in argv:
            m = extr.match(arg)
            if not m:
                raise ArgSyntaxError('Illegal argument format: "' + arg + '"')

            if m.group('key') is not None:
                key = m.group('key')
                val = m.group('val')
                if not hasattr(self, key):
                    raise ArgSyntaxError('Unknown argument: "' + key + '"')
                self._setVal(key,val)
                if key in keysToSet:
                    keysToSet.remove(key)
            else: # opts
                opts = m.group('opts')
                assert opts is not None
                for c in opts:
                    if not hasattr(self, c):
                        raise ArgSyntaxError('Unknown option: "' + c + '"')
                    exp = 'self.' + c + ' = True'
                    exec(exp)

        if not len(keysToSet) == 0:
            raise ArgSyntaxError('Missing arguments: %s' % str(list(keysToSet)))

        return self

    def _setVal(self, key, val):
        '''
        Takes a key-value pair of strings and converts them to
        a member variable with the key as member variable name, 
        and the val as the value. The type is resolved using the
        kwTypeMap dictionary.
        '''
                
        typ = self._kwTypeMap[key]
        if typ is list:
            val = '[' + val + ']'
        elif typ is bool:
            val = '"' + self._toBool(val) + '"'
        elif val == 'None':
            typ = NoneInitializer
            val = '"None"'
        elif val is None:
            raise ArgSyntaxError('Empty argument: "--' + key + '"')
        else:
            val = '"' + val + '"'

        exp = 'self.' + key + '=' + typ.__name__ + '(' + val + ')'
        exec(exp)
                
    def _toBool(self, val):
        if val in ['True', 'true', 'T', '1', 'OK', 'ok', 'Y', 'y', 'Yes', 'yes']:
            return '1'
        if val in ['False', 'false', 'F', '0', 'N', 'n', 'No', 'no']:
            return ''
        raise ArgSyntaxError('Illegal boolean token: "' + val + '"')

    def _initOpts(self, opts):
        for o in self._opts:
            exec('self.' + o + ' = False')
            

    def _initAttributes(self):
        for key in self._kwTypeMap:
            if key in self._defaults:
                self._setVal(key, str(self._defaults[key]))
            else:
                exp = 'self.' + key + ' = None'
                exec(exp)

    def usage(self):
        result = self.spec()
        if self._helpTexts is not None:
            result += '\nDetailed instructions:\n'
            for k in self._helpTexts:
                result += self._getKwdFormatLine(k) + ' :\n'
                result += self._getHelpText(k)
                result += '\n'*2
        return result

    def spec(self):
        result =  'Usage: [--<key>[:value]] [-opts]\n'

        result += 'Options:'
        for o in self._opts:
            result += ' -' + o
        result += '\n'

        result += 'Keys:\n'
        for k in self._kwTypeMap:
            result += self._getKwdFormatLine(k) + '\n'

        return result

    def _getKwdFormatLine(self,k):
        result = '   --' + k
        t = self._kwTypeMap[k]
        if t is not None:
            result += ':<' + t.__name__ + '>'
        if k in self._defaults:
            result += ' [default:' + str(self._defaults[k]) + ']'  
        return result

    def _getHelpText(self,k):
        lw  = 75
        ind = ' '*5
        raw = self._helpTexts[k]
        txt = ind
        ll  = 0
        for m in re.finditer(r'([^ \n]*)( |\n|$)', raw):
           tok,sep = (m.group(1),m.group(2))
           txt += tok
           ll  += len(tok)
           if sep == '\n':
               txt += '\n' + ind
               ll  = 0
           elif sep == ' ':
               if ll > lw:
                   txt += '\n' + ind
                   ll = 0
               else:
                   txt += ' '
                   ll  += 1

        return txt

    def __str__(self):
        result = 'args{'
        for o in self._opts:
            result += o + '=' + str(getattr(self,o)) + ','
            
        for k in self._kwTypeMap:
            result += k + '=' + str(getattr(self,k)) + ','

        return result[:-1] + '}'

    def __iter__(self):
        '''
        Support convert to dict
        '''
        for o in self._opts:
            yield o, getattr(self,o)
            
        for k in self._kwTypeMap:
            yield k, getattr(self,k)
        

def NoneInitializer(dah):
    assert dah == 'None'
    return None
    



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# Copyright 2020 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

from   ophac.dtypes import *
import ophac.ultrametric as ult


def _getLogger(x):
    import logging
    return logging.getLogger(__name__ + '.' + x.__name__)

class HAC:

    def __init__(self, lnk, ord=1, dK=1e-12, pred=True, cutoff=True):
        '''
        lnk  - Linkage model; one of 'single', 'average' or 'complete'.
        ord  - The norm-order to use. Default is 1.0.
        dK   - The increment to apply for the ultrametric completion.
        pred - Use partial order reduction? Default is true.
        cutoff - Use fit detorieration cutoff
        '''
        self.log    = _getLogger(HAC)
        self.lnk    = lnk
        self.ord    = ord
        self.dK     = dK
        self.pred   = pred
        self.cutoff = cutoff
        self._setLinkageFunctionFactory()
        self.log.info('Instantiated with lnk=%s and ord=%1.3f.', lnk, ord)

    def _setLinkageFunctionFactory(self):
        lnkFact = None
        if self.lnk == 'single':
            lnkFact = lambda sizes : BasicLinkage(SingleLinkage())
        elif self.lnk == 'complete':
            lnkFact = lambda sizes : BasicLinkage(CompleteLinkage())
        elif self.lnk == 'average':
            lnkFact = lambda sizes : BasicLinkage(AverageLinkage(sizes))
        else:
            raise Exception('Unknown linkage: "%s"' % self.lnk)

        self._getLinkageFunction = lnkFact

    def _initClustering(self, dissim, quivers):
        assert len(quivers) == dissim.n

        self.N  = len(quivers)
        self.d0 = dissim

        ac0  = AgglomerativeClustering()
        U0   = ult.ultrametric(ac0, self.N, self.dK)
        diff = (self.d0 - U0).norm(self.ord)

        self.bestDiff = diff       # Closest ultrametric
        self.acs      = [ac0]      # Maximal best chains
        self.ults     = set([U0])  # Visited ultrametrics

    def generate(self,dissim,order=None):
        '''
        dissim - Dissimilarity measure of type ophac.dtypes.DistMatrix 
        order  - Order relation of type ophac.dtypes.Quivers, 
                 or None for non-ordered clustering.
        '''
        if order is None:
            order = Quivers(n=dissim.n, relation=[])

        self._initClustering(dissim, order)

        P0   = Partition(n=self.N)
        
        self._exploreChains(dissim, order, P0, self.acs[-1], self.bestDiff)

        return _pickBest(dissim,self.acs,self.ord,self.dK)
        
    def _exploreChains(self, dissim, order, partition, ac0, lastDiff):
        '''
        dissim    - dissimilarity measure for the current partition
        order     - Quivers object for the current partition
        partition - Current partition
        ac0       - AgglomerativeClustering object leading up to the 
                    current partition
        '''
        self.log.debug('Exploring set of %d elements.', len(partition))

        if len(order) == 1:
            self.log.debug('Trivial partition reached.')
            self._registerCandidate(ac0)
            return 
    
        # Find minimal dissimilarity level to merge on
        chunks = dissim.getChunkedIndexPairs()
        chunk  = None
        for _ch in chunks[::-1]:
            if chunk is not None:
                break
            for a,b in _ch.pairs:
                if order.canMerge(a,b):
                    self.log.debug('Mergeable pair (%d,%d) of dissim %1.3f found.',
                                   a, b, _ch.dist)
                    chunk = _ch
                    break

        if chunk is None:
            self.log.debug('No further mergeable pairs --- maximal element reached')
            self._registerCandidate(ac0)
            return 

        # Get current linkage
        linker = self._getLinkageFunction([len(x) for x in partition])

        merged = False
        for a,b in chunk.pairs:
            if order.canMerge(a,b):
                self.log.debug('Merging (%d,%d) at dissim %1.3f.', a, b, chunk.dist)
                D2  = linker(a,b,dissim)
                P2  = partition.merge(a,b)
                O2  = order.merge(a,b)
                ac2 = ac0 + AC(joins=[(a,b)], dists=[chunk.dist])
                merged = True
                if len(ac2) > 1:
                    assert ac2.dists[-2] <= ac2.dists[-1]
                diff, better = self._checkVisit(ac2, lastDiff)
                if better:
                    self._exploreChains(D2, O2, P2, ac2, diff)

        assert merged

        return

    def _checkVisit(self, ac, lastDiff):
        '''
        Returns true if this chain should be attempted extended.
        '''
        U  = ult.ultrametric(ac, self.N, self.dK)
        
        if self.pred:
            l0 = len(self.ults)
            self.ults.add(U)
            l1 = len(self.ults)
            if l0 == l1:
                # Already visited this state
                return (None,False)

        if not self.cutoff:
            return (0,True)
        
        diff   = (U - self.d0).norm(self.ord)
        better = diff <= lastDiff or diff <= self.bestDiff
        return (diff, better)

    def _registerCandidate(self, ac):
        U    = ult.ultrametric(ac, self.N, self.dK)
        diff = (U - self.d0).norm(self.ord)

        if diff > self.bestDiff:
            # discard candidate
            return

        if diff < self.bestDiff:
            self.log.info('New best candidate at diff %1.3e', diff)
            self.acs      = [ac]
            self.bestDiff = diff
        else:
            # Same diff
            self.acs.append(ac)

def _pickBest(d0,acs,ord,dK):
    import numpy as np
    log = _getLogger(_pickBest)
    log.info('Finding best of %d dendrograms.', len(acs))

    N = d0.n
    
    if len(acs) == 1:
        log.info('Returning unique solution.')
        return acs

    # Remove duplicate ultrametric ACs
    acs2 = []
    ults = set()
    for ac in acs:
        U  = ult.ultrametric(ac, N, dK)
        l0 = len(ults)
        ults.add(U)
        l1 = len(ults)
        if l0 != l1:
            acs2.append(ac)
    acs = acs2

    norm = lambda ac : (d0 - ult.ultrametric(ac, N, dK)).norm(ord)
    acs  = sorted(acs, key=norm)
    best = norm(acs[0])
    i    = 1
    while i < len(acs) and norm(acs[i]) == best:
        i += 1

    log.info('Returning %d bests with fitting error %1.3f', i, norm(acs[0]))

    return acs[:i]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2025 Daniel Bakkelund
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension, build_ext
import os
import re
from glob import glob

def get_version():
    version_file = os.path.join("src", "ophac", "_version.py")
    with open(version_file) as f:
        content = f.read()
    match = re.search(r'__version__\s*=\s*["\']([^"\']+)["\']', content)
    if not match:
        raise RuntimeError("Unable to find version string in _version.py")
    return match.group(1)

ext_modules = [
    Pybind11Extension(
        "ophac_cpp",
        sources=sorted(glob("src/cpp/**/*.cpp", recursive=True)),
        include_dirs=["src/cpp"],
        cxx_std=17,
    )
]

setup(
    version=get_version(),
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)
